# Generated by Django 2.2.13 on 2020-06-30 12:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0011_rpmremote_sles_auth_token'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='packagecategory',
            name='packagegroups',
        ),
        migrations.RemoveField(
            model_name='packageenvironment',
            name='optionalgroups',
        ),
        migrations.RemoveField(
            model_name='packageenvironment',
            name='packagegroups',
        ),
        migrations.RemoveField(
            model_name='packagegroup',
            name='related_packages',
        ),
    ]
