# Generated by Django 2.2.16 on 2020-10-15 11:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0027_checksum_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmrepository',
            name='last_sync_repomd_checksum',
            field=models.CharField(max_length=64, null=True),
        ),
    ]
