# Generated by Django 2.2.14 on 2020-08-06 00:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0020_remove_updatecollection_m2m'),
    ]

    operations = [
        migrations.RenameField(
            model_name='updatecollection',
            old_name='_update_record',
            new_name='update_record',
        ),
    ]
