# Generated by Django 2.2.20 on 2021-05-12 11:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0030_DATA_fix_updaterecord'),
    ]

    operations = [
        migrations.AddField(
            model_name='modulemd',
            name='static_context',
            field=models.BooleanField(null=True),
        ),
    ]
