# Generated by Django 2.2.19 on 2021-04-02 21:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0033_new_distribution_model'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmrepository',
            name='autopublish',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='gpgcheck',
            field=models.IntegerField(choices=[(0, 0), (1, 1)], default=0),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='metadata_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='package_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='repo_gpgcheck',
            field=models.IntegerField(choices=[(0, 0), (1, 1)], default=0),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='sqlite_metadata',
            field=models.BooleanField(default=False),
        ),
    ]
